package ch.unige.aou.service;

import java.security.SecureRandom;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Base64;

import org.springframework.http.HttpHeaders;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.PathVariable;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.HashTool;

import ch.unige.aou.business.UserService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.security.User;
import ch.unige.aou.model.settings.DownloadToken;
import ch.unige.aou.repository.DownloadTokenRepository;

@Service
public class DownloadTokenService {

  private final UserService userService;
  private final DownloadTokenRepository downloadTokenRepository;
  private final SecureRandom secureRandom;
  private final int downloadTokenLifeTimeMinutes;

  // Token length should be a multiple of three to avoid padding with '=' character that could be
  // problematic in cookie value
  private static final int DOWNLOAD_TOKEN_LENGTH = 201;

  private static final long CLEANING_INTERVAL_MILLISECONDS = 600000; // ten minutes

  public DownloadTokenService(UserService userService, DownloadTokenRepository downloadTokenRepository, AouProperties aouProperties) {
    this.userService = userService;
    this.downloadTokenRepository = downloadTokenRepository;
    this.downloadTokenLifeTimeMinutes = aouProperties.getSecurity().getDownloadTokenLifeTimeMinutes();
    this.secureRandom = new SecureRandom();
  }

  @Scheduled(fixedDelay = CLEANING_INTERVAL_MILLISECONDS)
  public void removeExpiredDownloadToken() {
    final OffsetDateTime limitTime = OffsetDateTime.now().minusMinutes(this.downloadTokenLifeTimeMinutes);
    for (DownloadToken downloadToken : this.downloadTokenRepository.findExpiredDownloadTokens(limitTime)) {
      this.downloadTokenRepository.delete(downloadToken);
    }
  }

  public HttpHeaders getHeadersWithToken(@PathVariable String archiveId) {
    User user = this.userService.getCurrentUser();
    if (user == null) {
      throw new SolidifyRuntimeException("Cannot get authenticated user");
    }
    String userId = user.getResId();
    DownloadToken existingToken = this.downloadTokenRepository.findByArchiveIdAndUserId(archiveId, userId);
    if (existingToken != null) {
      this.downloadTokenRepository.delete(existingToken);
    }
    final byte[] bytes = new byte[DOWNLOAD_TOKEN_LENGTH];
    this.secureRandom.nextBytes(bytes);
    final String token = Base64.getEncoder().encodeToString(bytes);
    final String tokenHash = HashTool.hash(token);
    DownloadToken downloadToken = new DownloadToken(userId, archiveId, tokenHash, token);
    this.downloadTokenRepository.save(downloadToken);
    HttpHeaders headers = new HttpHeaders();
    OffsetDateTime expiresDate = OffsetDateTime.now().plusMinutes(this.downloadTokenLifeTimeMinutes);
    String expiresString = DateTimeFormatter.RFC_1123_DATE_TIME.format(expiresDate);
    headers.add(HttpHeaders.SET_COOKIE,
            "downloadToken-" + archiveId + "=" + token + "; Expires=" + expiresString + "; Path=/; Secure; HttpOnly");
    headers.add(HttpHeaders.ACCESS_CONTROL_ALLOW_CREDENTIALS, "true");
    return headers;
  }

}
