package ch.unige.aou.model.security;

import java.time.OffsetDateTime;
import java.util.Objects;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToOne;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import com.fasterxml.jackson.annotation.JsonFormat;
import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.auth.model.ApplicationRole;
import ch.unige.solidify.auth.model.UserInfo;
import ch.unige.solidify.message.CacheMessage;
import ch.unige.solidify.model.SolidifyApplicationRole;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.util.ReflectionTool;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.rest.ModuleName;

@Entity
public class User extends ResourceNormalized<User> implements UserInfo {
  private static final int LAST_LOGIN_IP_ADDRESS_SIZE = 255;

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_TIME_FORMAT)
  @Column(length = 3)
  private OffsetDateTime lastLoginTime;

  @Size(max = LAST_LOGIN_IP_ADDRESS_SIZE)
  private String lastLoginIpAddress;

  @NotNull
  @ManyToOne
  private SolidifyApplicationRole applicationRole;

  @NotNull
  private Boolean disabled = Boolean.FALSE;

  @NotNull
  @Size(min = 1)
  @Column(unique = true)
  private String email;

  @NotNull
  @Size(min = 1)
  @Column(unique = true)
  private String externalUid;

  @NotNull
  @Size(min = 1)
  private String firstName;

  @NotNull
  @Size(min = 1, max = 512)
  private String homeOrganization;

  @NotNull
  @Size(min = 1)
  private String lastName;

  @OneToOne(targetEntity = Person.class)
  @JoinColumn(name = AouConstants.DB_PERSON_ID, referencedColumnName = AouConstants.DB_RES_ID)
  private Person person;

  public User() {
  }

  public User(UserInfo userInfo) {
    this.setResId(userInfo.getResId());
    this.setExternalUid(userInfo.getExternalUid());
    this.setEmail(userInfo.getEmail());
    this.setApplicationRole(new SolidifyApplicationRole(userInfo.getApplicationRole()));
    this.setFirstName(userInfo.getFirstName());
    this.setLastName(userInfo.getLastName());
    this.setHomeOrganization(userInfo.getHomeOrganization());
    this.getCreation().setWhen(userInfo.getCreationWhen());
    this.getCreation().setWho(userInfo.getCreationWho());
    this.getLastUpdate().setWhen(userInfo.getLastUpdateWhen());
    this.getLastUpdate().setWho(userInfo.getLastUpdateWho());
  }

  @Override
  public ApplicationRole getApplicationRole() {
    return this.applicationRole;
  }

  @Override
  @Transient
  @JsonIgnore
  public CacheMessage getCacheMessage() {
    final CacheMessage cacheMessage = super.getCacheMessage();
    cacheMessage.addOtherProperty("externalUid", this.getExternalUid());
    return cacheMessage;
  }

  @Override
  public String getEmail() {
    return this.email;
  }

  @Override
  public String getExternalUid() {
    return this.externalUid;
  }

  @Override
  public String getFirstName() {
    return this.firstName;
  }

  public String getFullName() {
    if (this.getPerson() == null) {
      return this.getLastName() + ", " + this.getFirstName();
    }
    return this.getPerson().getFullName();
  }

  @Override
  public String getHomeOrganization() {
    return this.homeOrganization;
  }

  @Override
  public String getLastName() {
    return this.lastName;
  }

  public Person getPerson() {
    return this.person;
  }

  public String getLastLoginIpAddress() {
    return this.lastLoginIpAddress;
  }

  public OffsetDateTime getLastLoginTime() {
    return this.lastLoginTime;
  }

  @Transient
  @JsonIgnore
  public String getUniqueId() {
    return this.externalUid;
  }

  @Transient
  @JsonIgnore
  @Override
  public OffsetDateTime getCreationWhen() {
    return this.creation.getWhen();
  }

  @Transient
  @JsonIgnore
  @Override
  public OffsetDateTime getLastUpdateWhen() {
    return this.lastUpdate.getWhen();
  }

  @Transient
  @JsonIgnore
  @Override
  public String getCreationWho() {
    return this.creation.getWho();
  }

  @Transient
  @JsonIgnore
  @Override
  public String getLastUpdateWho() {
    return this.lastUpdate.getWho();
  }

  @Override
  public void init() {
    // Do nothing
  }

  @Transient
  @JsonIgnore
  public boolean isAccountNonExpired() {
    return !this.disabled;
  }

  @Transient
  @JsonIgnore
  public boolean isAccountNonLocked() {
    return this.isEnabled();
  }

  @Transient
  @JsonIgnore
  public boolean isCredentialsNonExpired() {
    return this.isAccountNonExpired();
  }

  @Override
  public boolean isEnabled() {
    return !this.disabled;
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public void setApplicationRole(SolidifyApplicationRole applicationRole) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), applicationRole);
  }

  public void setDisabled(Boolean disabled) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), disabled);
  }

  public void setEmail(String email) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), email);
  }

  public void setExternalUid(String externalUid) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), externalUid);
  }

  public void setFirstName(String firstName) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), firstName);
  }

  public void setHomeOrganization(String homeOrganization) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), homeOrganization);
  }

  public void setLastName(String lastName) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), lastName);
  }

  public void setPerson(Person person) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), person);
  }

  public void setLastLoginTime(OffsetDateTime newLoginTime) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), newLoginTime);
  }

  public void setLastLoginIpAddress(String newLoginIPAddress) {
    if (newLoginIPAddress != null && newLoginIPAddress.length() > LAST_LOGIN_IP_ADDRESS_SIZE) {
      newLoginIPAddress = newLoginIPAddress.substring(0, LAST_LOGIN_IP_ADDRESS_SIZE);
    }
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), newLoginIPAddress);
  }

  @Transient
  @JsonIgnore
  public void setUniqueId(String uniqueId) {
    this.setExternalUid(uniqueId);
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    builder.append("User [aai=").append(this.externalUid).append(", email=").append(this.email).append(", application role")
            .append(this.applicationRole).append("]");
    return builder.toString();
  }

  @Override
  public boolean equals(final Object other) {
    if (this == other) {
      return true;
    }
    if (!(other instanceof User)) {
      return false;
    }
    if (!super.equals(other)) {
      return false;
    }
    final User castOther = (User) other;
    return Objects.equals(this.applicationRole, castOther.applicationRole) && Objects.equals(this.email, castOther.email)
            && Objects.equals(this.externalUid, castOther.externalUid) && Objects.equals(this.firstName, castOther.firstName)
            && Objects.equals(this.homeOrganization, castOther.homeOrganization) && Objects.equals(this.lastName, castOther.lastName)
            && Objects.equals(this.person, castOther.person);
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.applicationRole, this.email, this.externalUid, this.firstName,
            this.homeOrganization, this.lastName, this.person);
  }

}
