package ch.unige.aou.business;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Optional;
import javax.transaction.Transactional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Service;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;

import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.model.Sortable;
import ch.unige.aou.model.settings.Label;
import ch.unige.aou.model.settings.License;
import ch.unige.aou.model.settings.LicenseGroup;
import ch.unige.aou.repository.LicenseRepository;
import ch.unige.aou.specification.LicenseSpecification;

@Service
public class LicenseService extends AouResourceService<License> {
  static final Logger log = LoggerFactory.getLogger(LicenseService.class);

  public static final String AOU_LICENSES_FILE = "aou-licenses.json";

  private LanguageService languageService;
  private LicenseGroupService licenseGroupService;

  public LicenseService(LanguageService languageService, LicenseGroupService licenseGroupService) {
    this.languageService = languageService;
    this.licenseGroupService = licenseGroupService;
  }

  public License findByOpenLicenseId(String openLicenseid) {
    return ((LicenseRepository) this.itemRepository).findByOpenLicenseId(openLicenseid);
  }

  public License findByTitle(String title) {
    return ((LicenseRepository) this.itemRepository).findByTitle(title);
  }

  public License findByUrl(URL url) {
    return ((LicenseRepository) this.itemRepository).findByUrl(url);
  }

  @Override
  public LicenseSpecification getSpecification(License resource) {
    return new LicenseSpecification(resource);
  }

  @Transactional
  public void initDefaultData() {
    try {
      ObjectMapper mapper = new ObjectMapper();
      final ArrayNode licensesArray = mapper
              .readValue(FileTool.toString(new ClassPathResource(AOU_LICENSES_FILE).getInputStream()), ArrayNode.class);
      for (final JsonNode item : licensesArray) {
        String resId = item.get("resId").textValue();
        if (!this.existsById(resId)) {
          License license = new License();
          license.setResId(resId);

          LicenseGroup licenseGroup = this.licenseGroupService.findOne(item.get("groupId").textValue());
          license.setLicenseGroup(licenseGroup);

          license.setOpenLicenseId(item.get("openLicenseId").textValue());
          license.setTitle(item.get("title").textValue());
          if (item.has("version")) {
            license.setVersion(item.get("version").textValue());
          }
          if (item.has("url")) {
            String url = item.get("url").textValue();
            try {
              license.setUrl(new URL(url));
            } catch (MalformedURLException e) {
              log.warn("Error in creating license: Url {} is malformed", url);
            }
          }
          license.setSortValue(item.get("sortValue").asInt());
          license.setVisible(true);

          if (item.has("labels")) {
            JsonNode labelsNode = item.get("labels");
            if (labelsNode.has(AouConstants.LANG_CODE_FRENCH)) {
              license.getLabels().add(new Label(labelsNode.get(AouConstants.LANG_CODE_FRENCH).textValue(),
                      this.languageService.findByCode(AouConstants.LANG_CODE_FRENCH)));
            }
            if (labelsNode.has(AouConstants.LANG_CODE_ENGLISH)) {
              license.getLabels().add(new Label(labelsNode.get(AouConstants.LANG_CODE_ENGLISH).textValue(),
                      this.languageService.findByCode(AouConstants.LANG_CODE_ENGLISH)));
            }
          }

          this.save(license);
          log.info("License '{}' created", resId);
        }
      }
    } catch (IOException e) {
      log.error("Unable to initialize licenses from Json file", e);
    }
  }

  @Override
  protected void beforeValidate(License item) {
    this.setDefaultSortValue(item);
  }

  @Override
  protected void validateItemSpecificRules(License license, BindingResult errors) {

    // check LicenseGroup
    if (license.getLicenseGroup() == null || StringTool.isNullOrEmpty(license.getLicenseGroup().getResId())) {
      errors.addError(new FieldError(license.getClass().getSimpleName(), "licenseGroup", "The license group is mandatory"));
    } else if (!this.licenseGroupService.existsById(license.getLicenseGroup().getResId())) {
      errors.addError(new FieldError(license.getClass().getSimpleName(), "licenseGroup", "The license group does not exist"));
    }

    // check title uniqueness
    License existingLicense = this.findByTitle(license.getTitle());
    if (existingLicense != null && !existingLicense.getResId().equals(license.getResId())) {
      errors.addError(new FieldError(license.getClass().getSimpleName(), "title", "This value already exists"));
    }

    // check open license id uniqueness
    existingLicense = this.findByOpenLicenseId(license.getOpenLicenseId());
    if (existingLicense != null && !existingLicense.getResId().equals(license.getResId())) {
      errors.addError(new FieldError(license.getClass().getSimpleName(), "openLicenseId", "This value already exists"));
    }
  }

  @Override
  public Optional<Integer> getMaxSortValue(Sortable item) {
    return ((LicenseRepository) this.itemRepository).findMaxSortValue();
  }
}
