package ch.unige.aou.business;

import java.time.OffsetDateTime;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.event.EventListener;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.message.EmailMessage;
import ch.unige.aou.model.notification.Event;
import ch.unige.aou.model.notification.EventMessage;
import ch.unige.aou.model.notification.EventType;
import ch.unige.aou.model.notification.Notification;
import ch.unige.aou.model.notification.NotificationType;
import ch.unige.aou.model.publication.Contributor;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.security.User;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.repository.NotificationRepository;
import ch.unige.aou.specification.NotificationSpecification;

@Service
public class NotificationService extends AouResourceService<Notification> {
  private static final Logger log = LoggerFactory.getLogger(NotificationService.class);

  private static final String MESSAGE_PUBLICATION_TO_VALIDATE = "notifications.message.new_deposit_to_validate";
  private static final String MESSAGE_MY_INDIRECT_PUBLICATION_VALIDATED = "notifications.message.indirect_deposit_validated";
  private static final String MESSAGE_MY_PUBLICATION_VALIDATED = "notifications.message.my_deposit_validated";
  private static final String MESSAGE_MY_PUBLICATION_REQUIRES_FEEDBACK = "notifications.message.deposit_requires_feeback";
  private static final String MESSAGE_MY_PUBLICATION_REJECTED = "notifications.message.my_deposit_rejected";
  private static final String MESSAGE_NEW_COMMENT_TO_VALIDATOR = "notifications.message.comment_deposit_validator";
  private static final String MESSAGE_NEW_COMMENT_TO_CONTRIBUTORS = "notifications.message.comment_deposit_contributor";
  private static final String MESSAGE_NEW_COMMENT_TO_DEPOSITOR = "notifications.message.comment_deposit_depositor";

  private final UserService userService;
  private final PersonService personService;
  private final NotificationTypeService notificationTypeService;
  private final PublicationService publicationService;
  private final EventService eventService;

  public NotificationService(UserService userService, PersonService personService, NotificationTypeService notificationTypeService,
          PublicationService publicationService, EventService eventService) {
    this.userService = userService;
    this.personService = personService;
    this.notificationTypeService = notificationTypeService;
    this.publicationService = publicationService;
    this.eventService = eventService;
  }

  @Override
  public NotificationSpecification getSpecification(Notification notification) {
    return new NotificationSpecification(notification);
  }

  /**
   * Use an async event to create notifications for a new Event as the treatment may be time consuming
   * <p>
   * Note: Transactional is required as the Event contains a Publication instance with lazy loaded structures
   * --> a JPA session is required to retrieve them
   *
   * @param event
   */
  @EventListener
  @Async
  @Transactional
  public void listenToNewEventCreation(EventMessage eventMessage) {

    Event event = this.eventService.findOne(eventMessage.getResId());

    switch (event.getEventType().getResId()) {
      case EventType.PUBLICATION_TO_VALIDATE_ID:
        this.createPublicationToValidateNotifications(event);
        break;
      case EventType.PUBLICATION_SUBMITTED_ID:
        this.createPublicationSubmittedNotifications(event);
        break;
      case EventType.PUBLICATION_FEEDBACK_REQUIRED_ID:
        this.createPublicationFeedbackRequiredNotification(event);
        break;
      case EventType.PUBLICATION_REJECTED_ID:
        this.createPublicationRejectedNotifications(event);
        break;
      case EventType.COMMENT_FOR_VALIDATORS_IN_PUBLICATION_ID:
        this.createValidatorCommentInPublicationNotification(event);
        break;
      case EventType.COMMENT_IN_PUBLICATION_ID:
        this.createValidatorCommentInPublicationNotification(event);
        this.createContributorsCommentInPublicationNotification(event);
        this.createDepositorCommentInPublicationNotification(event);
        break;
      default:
        break;
    }
  }

  /**
   * Create a notification for the depositor when a publication is rejected
   *
   * @param event The new event for which notifications must be created
   */
  private void createPublicationRejectedNotifications(Event event) {
    log.info("will create notifications for new Publications rejected");

    this.createNotification(event.getPublication().getCreator(), NotificationType.MY_PUBLICATION_REJECTED.getResId(), event,
            this.messageService.get(MESSAGE_MY_PUBLICATION_REJECTED,
                    new Object[] { event.getPublication().getLastComment(false).getText(),
                            event.getPublication().getLastComment(false).getPerson().getFullName() }));

    log.info("notifications for new Publications rejected created");
  }

  /**
   * Create a notification for all users that can validate this publication
   *
   * @param event The new event for which notifications must be created
   */
  public void createPublicationToValidateNotifications(Event event) {
    log.info("will create notifications for new Publications to validate");

    List<Person> validators = this.personService.getValidators(event.getPublication());

    for (Person validator : validators) {
      this.createNotification(validator, NotificationType.PUBLICATION_TO_VALIDATE.getResId(), event,
              this.messageService.get(MESSAGE_PUBLICATION_TO_VALIDATE, new Object[] { event.getPublication().getTitle() }));
    }

    log.info("notifications for new Publications to validate created");
  }

  /**
   * Create a notification for the depositor and for all UNIGE contributors (the ones having a cnIndividu)
   * when a Publication gets validated
   *
   * @param event
   */
  public void createPublicationSubmittedNotifications(Event event) {
    this.createPublicationSubmittedNotificationForDepositor(event);
    this.createPublicationSubmittedNotificationsForContributors(event);
    this.markAsReadPreviousRelatedNotifications(event.getPublication());
  }

  private void markAsReadPreviousRelatedNotifications(Publication publication) {
   List<Notification> notificationsList = ((NotificationRepository)this.itemRepository).findByPublicationId(publication.getResId());

   //mark as read all notifications that are not from type MY_PUBLICATION_VALIDATED and MY_INDIRECT_PUBLICATION_VALIDATED
    notificationsList.stream().filter(n-> !n.getNotificationType().equals(NotificationType.MY_INDIRECT_PUBLICATION_VALIDATED) &&
            !n.getNotificationType().equals(NotificationType.MY_PUBLICATION_VALIDATED))
            .forEach(n -> {
              n.setReadTime(OffsetDateTime.now());
              this.save(n);
            });
  }

  /**
   * Update the sentTime property of a Notification
   * Note: Transactional is required to allow the method to be called from a scheduled task context that runs outside of the main thread JPA session
   *
   * @param notificationId
   * @param sentTime
   * @return
   */
  @Transactional
  public Notification updateSentTime(String notificationId, OffsetDateTime sentTime) {
    Notification notif = this.findOne(notificationId);
    notif.setSentTime(sentTime);
    return this.save(notif);
  }

  private void createPublicationSubmittedNotificationForDepositor(Event event) {
    log.info("will create notifications for validated Publications to depositors");

    this.createNotification(event.getPublication().getCreator(), NotificationType.MY_PUBLICATION_VALIDATED.getResId(), event,
            this.messageService.get(MESSAGE_MY_PUBLICATION_VALIDATED, new Object[] { event.getPublication().getTitle() }));

    log.info("notifications for validated Publications to depositors created");
  }

  private void createValidatorCommentInPublicationNotification(Event event) {
    log.info("will create notifications to validators for a new comment on Publication '{}'", event.getPublication().getResId());

    // useful to prevent a lazy loading exception on publicationStructures when calling `this.personService.getValidators(publication)`
    Publication publication = this.publicationService.findOne(event.getPublication().getResId());

    List<Person> validators = this.personService.getValidators(publication);

    for (Person validator : validators) {
      this.createNotification(validator, NotificationType.PUBLICATION_TO_VALIDATE_COMMENTED.getResId(), event,
              this.messageService.get(MESSAGE_NEW_COMMENT_TO_VALIDATOR,
                      new Object[] { event.getPublication().getLastComment(true).getText(),
                              event.getPublication().getLastComment(true).getPerson().getFullName() }));
    }

    log.info("notifications for new validators comments on publication created");
  }

  private void createContributorsCommentInPublicationNotification(Event event) {
    log.info("will create notifications to contributors for a new comment on Publication '{}'", event.getPublication().getResId());

    String message = this.messageService.get(MESSAGE_NEW_COMMENT_TO_CONTRIBUTORS,
            new Object[] { event.getPublication().getLastComment(false).getText(),
                    event.getPublication().getLastComment(false).getPerson().getFullName() });
    this.createNotificationForUnigeContributors(event, NotificationType.MY_INDIRECT_PUBLICATION_COMMENTED, message);

    log.info("notifications to contributors for a new comment on Publication created");
  }

  private void createDepositorCommentInPublicationNotification(Event event) {
    log.info("will create notification to depositor for a new comment on Publication '{}'", event.getPublication().getResId());

    Publication publication = this.publicationService.findOne(event.getPublication().getResId());

    this.createNotification(publication.getCreator(), NotificationType.MY_PUBLICATION_COMMENTED.getResId(), event,
            this.messageService.get(MESSAGE_NEW_COMMENT_TO_DEPOSITOR,
                    new Object[] { event.getPublication().getLastComment(false).getText(),
                            event.getPublication().getLastComment(false).getPerson().getFullName() }));

    log.info("notification to depositor for a new comment on Publication created");
  }

  private void createPublicationSubmittedNotificationsForContributors(Event event) {
    log.info("will create notifications to UNIGE contributors for the validated Publication '{}'", event.getPublication().getResId());

    String message = this.messageService.get(MESSAGE_MY_INDIRECT_PUBLICATION_VALIDATED, new Object[] { event.getPublication().getTitle() });
    this.createNotificationForUnigeContributors(event, NotificationType.MY_INDIRECT_PUBLICATION_VALIDATED, message);

    log.info("notifications to UNIGE contributors for the validated Publication '{}' created", event.getPublication().getResId());
  }

  private void createPublicationFeedbackRequiredNotification(Event event) {
    log.info("will create notification for publication requiring feedback");

    this.createNotification(event.getPublication().getCreator(), NotificationType.MY_PUBLICATION_FEEDBACK_REQUIRED.getResId(), event,
            this.messageService.get(MESSAGE_MY_PUBLICATION_REQUIRES_FEEDBACK,
                    new Object[] { event.getPublication().getLastComment(false).getText(),
                            event.getPublication().getLastComment(false).getPerson().getFullName() }));

    log.info("notification created for publication requiring feedback");

    this.sendPublicationEmailToDepositor(event.getPublication(), EmailMessage.EmailTemplate.MY_DEPOSIT_REQUIRES_FEEDBACK);
  }

  private void sendPublicationEmailToDepositor(Publication publication, EmailMessage.EmailTemplate template) {
    User recipientUser = this.userService.findByPersonResId(publication.getCreator().getResId());
    String[] parameters = new String[] { publication.getResId() };

    SolidifyEventPublisher.getPublisher()
            .publishEvent(new EmailMessage(recipientUser.getEmail(), template, parameters));
  }

  private void createNotification(Person person, String notificationTypeId, Event event, String message) {

    NotificationType notificationType = this.notificationTypeService.findOne(notificationTypeId);

    Notification notification = new Notification();
    notification.setRecipient(person);
    notification.setNotificationType(notificationType);
    notification.setEvent(event);
    notification.setMessage(message);
    this.save(notification);
  }

  private void createNotificationForUnigeContributors(Event event, NotificationType notificationType, String message) {
    // get contributors that are members of UNIGE
    List<Contributor> unigeContributors = event.getPublication().getContributors().stream()
            .filter(contributor -> !StringTool.isNullOrEmpty(contributor.getCnIndividu())).collect(Collectors.toList());

    //for each, check if they are user of the application
    for (Contributor unigeContributor : unigeContributors) {
      try {
        User user = this.userService.findByExternalUid(this.cnIndividuToExternalUid(unigeContributor.getCnIndividu()));
        this.createNotification(user.getPerson(), notificationType.getResId(), event, message);
      } catch (NoSuchElementException e) {
        log.debug("No user with cnIndividu {} found", unigeContributor.getCnIndividu());
      }
    }
  }
}
