package ch.unige.aou.service.metadata;

import java.math.BigInteger;
import java.time.LocalDate;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.DateTimeParseException;
import java.time.temporal.ChronoField;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.validation.ValidationError;

import ch.unige.aou.AouMetadataVersion;
import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.LanguageService;
import ch.unige.aou.business.PersonService;
import ch.unige.aou.business.PublicationSubSubtypeService;
import ch.unige.aou.business.PublicationSubtypeService;
import ch.unige.aou.business.PublicationTypeService;
import ch.unige.aou.business.ResearchGroupService;
import ch.unige.aou.business.StructureService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.display.StructureValidationRight;
import ch.unige.aou.model.publication.AbstractContributor;
import ch.unige.aou.model.publication.Contributor;
import ch.unige.aou.model.publication.ContributorDTO;
import ch.unige.aou.model.publication.MetadataCollection;
import ch.unige.aou.model.publication.MetadataCorrection;
import ch.unige.aou.model.publication.MetadataDates;
import ch.unige.aou.model.publication.MetadataFile;
import ch.unige.aou.model.publication.MetadataFunding;
import ch.unige.aou.model.publication.MetadataLink;
import ch.unige.aou.model.publication.MetadataTextLang;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.publication.PublicationStructure;
import ch.unige.aou.model.publication.PublicationSubSubtype;
import ch.unige.aou.model.publication.PublicationSubtype;
import ch.unige.aou.model.publication.PublicationType;
import ch.unige.aou.model.rest.JournalTitleDTO;
import ch.unige.aou.model.settings.Language;
import ch.unige.aou.model.settings.ResearchGroup;
import ch.unige.aou.model.settings.Structure;
import ch.unige.aou.model.xml.deposit.v2_4.AuthorRole;
import ch.unige.aou.service.rest.JournalTitleRemoteService;

public abstract class MetadataExtractor {

  private static final Logger log = LoggerFactory.getLogger(MetadataExtractor.class);

  private static final String INCORRECT_DATE = "yyyy-mm";

  private static final String ERROR_VALIDATION_STRUCTURE_NOT_AUTOMATICALLY_SET = "deposit.error.structure.validation_structure_not_automatically_set";

  protected PublicationTypeService publicationTypeService;
  protected PublicationSubtypeService publicationSubtypeService;
  protected PublicationSubSubtypeService publicationSubSubtypeService;
  protected StructureService structureService;
  protected ResearchGroupService researchGroupService;
  protected ContributorService contributorService;
  protected LanguageService languageService;
  protected JournalTitleRemoteService journalTitleRemoteService;
  protected PersonService personService;
  protected final MessageService messageService;

  private String[] metadataOutputDateFormat;
  private String[] frontendEnterDateFormat;
  private String[] frontendOutputDateFormat;
  private Map<String, String> defaultContributorRolesByDepositSubtype;

  protected MetadataExtractor(AouProperties aouProperties, PublicationTypeService publicationTypeService,
          PublicationSubtypeService publicationSubtypeService,
          PublicationSubSubtypeService publicationSubSubtypeService,
          StructureService structureService,
          ResearchGroupService researchGroupService,
          ContributorService contributorService,
          LanguageService languageService,
          JournalTitleRemoteService journalTitleRemoteService, PersonService personService, MessageService messageService) {
    this.publicationTypeService = publicationTypeService;
    this.publicationSubtypeService = publicationSubtypeService;
    this.publicationSubSubtypeService = publicationSubSubtypeService;
    this.structureService = structureService;
    this.researchGroupService = researchGroupService;
    this.contributorService = contributorService;
    this.languageService = languageService;
    this.journalTitleRemoteService = journalTitleRemoteService;
    this.personService = personService;
    this.messageService = messageService;

    this.metadataOutputDateFormat = aouProperties.getParameters().getMetadataOutputDateFormat();
    this.frontendEnterDateFormat = aouProperties.getParameters().getFrontendEnterDateFormat();
    this.frontendOutputDateFormat = aouProperties.getParameters().getFrontendOutputDateFormat();
    this.defaultContributorRolesByDepositSubtype = aouProperties.getMetadata().getDefault().getContributorRolesByDepositSubtype();
  }

  /*******************************************************************************************/

  public void setMetadataFromFormData(Publication publication) {
    String xmlMetadata = this.transformFormDataToMetadata(publication.getFormData());
    publication.setMetadata(xmlMetadata);
    publication.setMetadataVersion(AouMetadataVersion.getDefaultVersion());
  }

  public void setFormDataFromMetadata(Publication publication) {
    String formData = this.transformMetadataToFormData(publication.getMetadata());
    publication.setFormData(formData);
  }

  /**
   * Replace ids values by their text value (useful to be stored in pivotal metadata)
   *
   * @param depositDocObj Pivotal metadata in memory object
   */
  public void replaceIdsByTextValues(Object depositDocObj) {

    /*
     * Type
     */
    String typeStr = this.getType(depositDocObj);
    if (!StringTool.isNullOrEmpty(typeStr)) {
      Optional<PublicationType> type = this.publicationTypeService.findById(typeStr);
      type.ifPresent(publicationType -> this.setType(depositDocObj, publicationType.getName()));
    }

    /*
     * Subtype
     */
    String subtypeStr = this.getSubtype(depositDocObj);
    if (!StringTool.isNullOrEmpty(subtypeStr)) {
      Optional<PublicationSubtype> subType = this.publicationSubtypeService.findById(subtypeStr);
      subType.ifPresent(publicationSubtype -> this.setSubtype(depositDocObj, publicationSubtype.getName()));
    }

    /*
     * SubSubtype
     */
    String subSubtypeStr = this.getSubSubtype(depositDocObj);
    if (!StringTool.isNullOrEmpty(subSubtypeStr)) {
      Optional<PublicationSubSubtype> subSubType = this.publicationSubSubtypeService.findById(subSubtypeStr);
      subSubType.ifPresent(publicationSubSubtype -> this.setSubSubtype(depositDocObj, publicationSubSubtype.getName()));
    }

    /*
     * Title's language
     */
    String titleLangStr = this.getTitleLang(depositDocObj);
    if (!StringTool.isNullOrEmpty(titleLangStr)) {
      Optional<Language> language = this.languageService.findById(titleLangStr);
      language.ifPresent(lang -> this.setTitleLang(depositDocObj, lang.getCode()));
    }

    /*
     * Original Title's language
     */
    String originalTitleLangStr = this.getOriginalTitleLang(depositDocObj);
    if (!StringTool.isNullOrEmpty(originalTitleLangStr)) {
      Optional<Language> language = this.languageService.findById(originalTitleLangStr);
      language.ifPresent(lang -> this.setOriginalTitleLang(depositDocObj, lang.getCode()));
    }

    /*
     * Abstracts' languages
     */
    List<MetadataTextLang> abstracts = this.getAbstracts(depositDocObj);
    if (!abstracts.isEmpty()) {
      for (MetadataTextLang abstractText : abstracts) {
        if (!StringTool.isNullOrEmpty(abstractText.getLang())) {
          Optional<Language> language = this.languageService.findById(abstractText.getLang());
          language.ifPresent(lang -> abstractText.setLang(lang.getCode()));
        }
      }
      this.setAbstracts(depositDocObj, abstracts);
    }

    /*
     * Links descriptions language
     */
    List<MetadataLink> links = this.getLinks(depositDocObj);
    if (!links.isEmpty()) {
      for (MetadataLink link : links) {
        if (link.getDescription() != null && !StringTool.isNullOrEmpty(link.getDescription().getLang())) {
          Optional<Language> language = this.languageService.findById(link.getDescription().getLang());
          language.ifPresent(lang -> link.getDescription().setLang(lang.getCode()));
        }
      }
      this.setLinks(depositDocObj, links);
    }
  }

  /**
   * Replace text values by their ids (useful for the frontend form that uses ids for select values)
   *
   * @param depositDocObj Pivotal metadata in memory object
   */
  public void replaceTextValuesByIds(Object depositDocObj) {

    /*
     * Type
     */
    String typeStr = this.getType(depositDocObj);
    if (!StringTool.isNullOrEmpty(typeStr)) {
      PublicationType type = this.publicationTypeService.findByName(typeStr);
      if (type != null) {
        this.setType(depositDocObj, type.getResId());
      }
    }

    /*
     * Subtype
     */
    String subtypeStr = this.getSubtype(depositDocObj);
    if (!StringTool.isNullOrEmpty(subtypeStr)) {
      Optional<PublicationSubtype> subtypeOpt = this.publicationSubtypeService.findByName(subtypeStr);
      if (subtypeOpt.isPresent()) {
        this.setSubtype(depositDocObj, subtypeOpt.get().getResId());

        /*
         * SubSubtype
         */
        String subSubtypeStr = this.getSubSubtype(depositDocObj);
        if (!StringTool.isNullOrEmpty(subSubtypeStr)) {
          PublicationSubSubtype subSubtype = this.publicationSubSubtypeService.findByNameAndPublicationSubType(subSubtypeStr, subtypeOpt.get());
          if (subSubtype != null) {
            this.setSubSubtype(depositDocObj, subSubtype.getResId());
          }
        }
      }
    }

    /*
     * Title's language
     */
    String titleLangStr = this.getTitleLang(depositDocObj);
    if (!StringTool.isNullOrEmpty(titleLangStr)) {
      Language language = this.languageService.findByCode(titleLangStr);
      if (language != null) {
        this.setTitleLang(depositDocObj, language.getResId());
      }
    }

    /*
     * Original title's language
     */
    String originalTitleLangStr = this.getOriginalTitleLang(depositDocObj);
    if (!StringTool.isNullOrEmpty(originalTitleLangStr)) {
      Language language = this.languageService.findByCode(originalTitleLangStr);
      if (language != null) {
        this.setOriginalTitleLang(depositDocObj, language.getResId());
      }
    }

    /*
     * Abstracts' languages
     */
    List<MetadataTextLang> abstracts = this.getAbstracts(depositDocObj);
    if (!abstracts.isEmpty()) {
      for (MetadataTextLang abstractText : abstracts) {
        if (!StringTool.isNullOrEmpty(abstractText.getLang())) {
          Language language = this.languageService.findByCode(abstractText.getLang());
          if (language != null) {
            abstractText.setLang(language.getResId());
          }
        }
      }
      this.setAbstracts(depositDocObj, abstracts);
    }

    /*
     * Links descriptions language
     */
    List<MetadataLink> links = this.getLinks(depositDocObj);
    if (!links.isEmpty()) {
      for (MetadataLink link : links) {
        if (link.getDescription() != null && !StringTool.isNullOrEmpty(link.getDescription().getLang())) {
          Language language = this.languageService.findByCode(link.getDescription().getLang());
          if (language != null) {
            link.getDescription().setLang(language.getResId());
          }
        }
      }
      this.setLinks(depositDocObj, links);
    }
  }

  /**
   * Replace fields values that are different between front form data and pivotal metadata
   *
   * @param depositDocObj
   */
  public void replaceFormValuesByMetadataValues(Object depositDocObj) {
    this.replaceIdsByTextValues(depositDocObj);
    this.formatDepositDocDatesToMetadataFormat(depositDocObj);
  }

  /**
   * Replace fields values that are different between front form data and pivotal metadata
   *
   * @param depositDocObj
   */
  public void replaceMetadataValuesByFormValues(Object depositDocObj) {
    this.replaceTextValuesByIds(depositDocObj);
    this.formatDepositDocDatesToFrontendFormat(depositDocObj);
  }

  public void replaceISSNByISSNL(Publication publication) {
    Object depositDocObj = this.createDepositDocObjectFromXml(publication.getMetadata());

    String issn = this.getISSN(depositDocObj);
    if (!StringTool.isNullOrEmpty(issn)) {
      //Find issn-l related with the issn using journalService
      ch.unige.solidify.rest.Collection<JournalTitleDTO> journalTitles = this.journalTitleRemoteService.getJournalTitlesByISSN(issn);
      if (journalTitles.get_data() != null && journalTitles.get_data().iterator().hasNext()) {
        JournalTitleDTO journalTitleDTO = journalTitles.get_data().iterator().next();
        String issnl = journalTitleDTO.getIssnl();
        if (!issn.equals(issnl)) {
          this.setISSN(depositDocObj, issnl);
        }
      }
      String metadata = this.serializeDepositDocToXml(depositDocObj);
      publication.setMetadata(metadata);
    }
  }

  /**
   * Set the publication properties that exists in metadata with the values in metadata
   *
   * @param publication
   */
  public void setPropertiesFromMetadata(Publication publication) {

    Object depositDocObj = this.createDepositDocObjectFromXml(publication.getMetadata());

    // Title
    String titleContent = this.getTitleContent(depositDocObj);
    if (!StringTool.isNullOrEmpty(titleContent)) {
      publication.setTitle(titleContent);
    }

    // Subtype
    PublicationSubtype subtype = null;
    String subtypeStr = this.getSubtype(depositDocObj);
    if (!StringTool.isNullOrEmpty(subtypeStr)) {
      Optional<PublicationSubtype> subtypeOpt = this.publicationSubtypeService.findByName(subtypeStr);
      if (subtypeOpt.isPresent()) {
        subtype = subtypeOpt.get();
        publication.setSubtype(subtype);
      }
    }

    // Subsubtype
    String subSubtypeStr = this.getSubSubtype(depositDocObj);
    if (subtype != null && !StringTool.isNullOrEmpty(subSubtypeStr)) {
      PublicationSubSubtype subSubtype = this.publicationSubSubtypeService.findByNameAndPublicationSubType(subSubtypeStr, subtype);
      if (subSubtype != null) {
        publication.setSubSubtype(subSubtype);
      }
    } else {
      publication.setSubSubtype(null);
    }

    // DOI
    String doi = this.getDOI(depositDocObj);
    if (!StringTool.isNullOrEmpty(doi)) {
      publication.setDoi(doi);
    } else {
      publication.setDoi(null);
    }

    // PMID
    BigInteger pmid = this.getPmid(depositDocObj);
    if (pmid != null) {
      publication.setPmid(pmid.toString());
    } else {
      publication.setPmid(null);
    }
  }

  public void setRelationsFromMetadata(Publication publication) {

    Object depositDocObj = this.createDepositDocObjectFromXml(publication.getMetadata());

    /*
     * Structures
     */
    List<Structure> structures = this.getStructures(depositDocObj);
    if (structures != null) {
      this.syncPublicationStructuresList(publication, structures);
    }

    /*
     * Research groups
     */
    List<ResearchGroup> groups = this.getResearchGroups(depositDocObj);
    if (groups != null) {
      this.syncResearchGroupsList(publication, groups);
    }

    /*
     * Contributors
     */
    List<AbstractContributor> abstractContributors = this.getContributors(depositDocObj);
    if (abstractContributors != null) {
      this.syncContributorsList(publication, abstractContributors);
    }
  }

  public JSONArray transformAcademicStructuresToJson(List<Structure> structures) {
    JSONArray structuresArray = new JSONArray();
    for (Structure st : structures) {
      structuresArray.put(new JSONObject()
              .put("name", st.getName())
              .put("code", st.getCodeStruct())
              .put("oldCode", st.getCnStructC())
              .put("resId", st.getResId())
      );
    }
    return structuresArray;
  }

  public JSONArray transformGroupsToJson(List<ResearchGroup> researchGroups) {
    JSONArray groupsArray = new JSONArray();
    for (ResearchGroup rg : researchGroups) {
      groupsArray.put(new JSONObject()
              .put("name", rg.getName())
              .put("code", rg.getCode())
              .put("resId", rg.getResId())
      );
    }
    return groupsArray;
  }

  public void setDefaultContributorsRoles(Object depositDocObj) {

    String subtype = this.getSubtype(depositDocObj);
    Optional<PublicationSubtype> publicationSubtypeOpt = this.publicationSubtypeService.findByName(subtype);
    if (publicationSubtypeOpt.isPresent()) {
      AuthorRole defaultRole = AuthorRole.fromValue(this.defaultContributorRolesByDepositSubtype.get(publicationSubtypeOpt.get().getResId()));

      List<AbstractContributor> contributors = this.getContributors(depositDocObj);

      contributors.stream()
              .filter(c -> c instanceof ContributorDTO && StringTool.isNullOrEmpty(((ContributorDTO) c).getRole()))
              .map(ContributorDTO.class::cast)
              .forEach(c -> c.setRole(defaultRole.value()));

      this.setContributors(depositDocObj, contributors);
    } else {
      throw new SolidifyRuntimeException(String.format("PublicationSubtype '%s' not found", subtype));
    }
  }

  /*******************************************************************************************/

  /**
   * Transform metadata from the format used by the frontend form to the format used to store metadata
   *
   * @param formData
   * @return
   */
  public abstract String transformFormDataToMetadata(String formData);

  /**
   * Transform metadata from the format used to store metadata to the format used by the frontend form
   *
   * @param metadata
   * @return
   */
  public abstract String transformMetadataToFormData(String metadata);

  /**
   * Serialize a object representing the pivotal metadata to string
   *
   * @param depositDocObj
   * @return
   */
  public abstract String serializeDepositDocToXml(Object depositDocObj);

  public abstract void completeMetadataWithFiles(Publication publication);

  public abstract void completeClassificationsMetadataWithDeweyCodes(Publication publication);

  protected abstract void formatDepositDocDatesToFrontendFormat(Object depositDocObj);

  protected abstract void formatDepositDocDatesToMetadataFormat(Object depositDocObj);

  public abstract Object createDepositDocObjectFromXml(String xmlData);

  public abstract Object createDepositDocFromJsonFormData(String jsonData);

  public abstract String getPmcIdFromPublication(Publication publication);

  public abstract void addResearchGroup(ResearchGroup researchGroup, Publication publication);

  public abstract void addStructure(Structure structure, Publication publication);

  public abstract void addLanguage(Language language, Publication publication);

  public abstract boolean isCnIndividuInContributors(Publication publication, String cnIndividu);

  public abstract void removeCnIndividuFromContributors(Publication publication, String cnIndividu);

  public abstract void cleanMetadata(Publication publication);

  public abstract void cleanMetadataToClone(Publication publication);

  /*******************************************************************************************/

  abstract String getType(Object depositDocObj);

  abstract void setType(Object depositDocObj, String value);

  abstract String getSubtype(Object depositDocObj);

  abstract void setSubtype(Object depositDocObj, String value);

  abstract String getSubSubtype(Object depositDocObj);

  abstract void setSubSubtype(Object depositDocObj, String value);

  abstract String getTitleContent(Object depositDocObj);

  abstract String getTitleLang(Object depositDocObj);

  abstract void setTitleLang(Object depositDocObj, String value);

  abstract String getOriginalTitleContent(Object depositDocObj);

  abstract String getOriginalTitleLang(Object depositDocObj);

  abstract void setOriginalTitleLang(Object depositDocObj, String value);

  abstract String getISBN(Object depositDocObj);

  abstract String getISSN(Object depositDocObj);

  abstract void setISSN(Object depositDocObj, String value);

  abstract String getDOI(Object depositDocObj);

  abstract BigInteger getPmid(Object depositDocObj);

  abstract String getArxivId(Object depositDocObj);

  abstract String getURN(Object depositDocObj);

  abstract String getPmcId(Object depositDocObj);

  public abstract List<AbstractContributor> getContributors(Object depositDocObj);

  public abstract void setContributors(Object depositDocObj, List<AbstractContributor> abstractContributors);

  abstract List<ResearchGroup> getResearchGroups(Object depositDocObj);

  abstract List<Structure> getStructures(Object depositDocObj);

  abstract List<MetadataDates> getDates(Object depositDocObj);

  abstract List<MetadataTextLang> getAbstracts(Object depositDocObj);

  abstract void setAbstracts(Object depositDocObj, List<MetadataTextLang> abstracts);

  abstract List<MetadataLink> getLinks(Object depositDocObj);

  abstract void setLinks(Object depositDocObj, List<MetadataLink> links);

  abstract List<MetadataCollection> getCollections(Object depositDocObj);

  abstract List<String> getLanguages(Object depositDocObj);

  abstract List<MetadataFunding> getFundings(Object depositDocObj);

  abstract List<String> getDatasets(Object depositDocObj);

  abstract List<MetadataFile> getFiles(Object depositDocObj);

  abstract List<MetadataCorrection> getCorrections(Object depositDocObj);

  /*******************************************************************************************/

  /**
   * Manage links between a publication and their structures according to the given list of academic structures: add structures that exist in
   * the academic structures list but not in the publication's structures list. Remove structures from the publication's structures that are not
   * in the academic structures list.
   *
   * @param publication
   * @param academicStructures
   */
  public void syncPublicationStructuresList(Publication publication, List<Structure> academicStructures) {
    List<PublicationStructure> structures = publication.getPublicationStructures();

    List<String> existingResIds;
    if (academicStructures.isEmpty()) {
      // if there is no structure in metadata, do not remove existing validation structure
      existingResIds = structures.stream().filter(pubStruct -> pubStruct.getLinkType() == PublicationStructure.LinkType.METADATA)
              .map(pubStruct -> pubStruct.getStructure().getResId()).collect(Collectors.toList());
    } else {
      // if there is one metadata structure, validation structure may be cleared as it is not useful anymore
      existingResIds = structures.stream().map(pubStruct -> pubStruct.getStructure().getResId()).collect(Collectors.toList());
    }

    List<String> resIdsToPersist = academicStructures.stream().map(ResourceNormalized::getResId).collect(Collectors.toList());
    List<String> resIdsToAdd = resIdsToPersist.stream().filter(element -> !existingResIds.contains(element)).collect(Collectors.toList());
    List<String> resIdsToRemove = existingResIds.stream().filter(element -> !resIdsToPersist.contains(element)).collect(Collectors.toList());
    List<PublicationStructure> publicationStructuresToRemove = structures.stream()
            .filter(pubStruct -> resIdsToRemove.contains(pubStruct.getStructure().getResId())).collect(Collectors.toList());

    // remove structures that are not present anymore
    structures.removeAll(publicationStructuresToRemove);

    // add structures not in list yet
    for (String resIdToAdd : resIdsToAdd) {
      // Add new Structure
      Structure structure = this.structureService.findById(resIdToAdd).orElse(null);
      if (structure != null) {
        PublicationStructure pStructure = new PublicationStructure();
        pStructure.setStructure(structure);
        pStructure.setPublication(publication);
        pStructure.setLinkType(PublicationStructure.LinkType.METADATA);

        structures.add(pStructure);
      } else {
        log.warn("No structure with resId '{}' found", resIdToAdd);
      }
    }

    // Manage specific case of an existing VALIDATION structure that is replaced by the same structure in METADATA
    // In this particular case, the logic above preserve the PublicationStructure in the structures list so its linkType must be updated
    for (PublicationStructure publicationStructure : structures) {
      if (publicationStructure.getLinkType() == PublicationStructure.LinkType.VALIDATION && resIdsToPersist.contains(
              publicationStructure.getStructure().getResId())) {
        publicationStructure.setLinkType(PublicationStructure.LinkType.METADATA);
        log.info("link type between structure '{}' and publication '{}' automatically updated to METADATA",
                publicationStructure.getStructure().getName(), publication.getResId());
      }
    }

    // If there isn't any structure anymore on the publication, it disappears from validator's list
    // --> in this case we take the first current user structure on which he has validation rights and we use it as VALIDATION structure
    if (structures.isEmpty() && publication.getStatus() == Publication.PublicationStatus.IN_VALIDATION) {
      if (!this.addCurrentUserFirstValidationStructureForSubtype(publication, structures)) {

        // We build the validation error here (instead of MetadataValidatorService) because validation has already been done at this step
        final BindingResult errors = new BeanPropertyBindingResult(publication, publication.getClass().getName());
        errors.addError(new FieldError(publication.getClass().getSimpleName(), "structure",
                this.messageService.get(ERROR_VALIDATION_STRUCTURE_NOT_AUTOMATICALLY_SET)));
        throw new SolidifyValidationException(new ValidationError(errors));
      }
    }
  }

  private boolean addCurrentUserFirstValidationStructureForSubtype(Publication publication, List<PublicationStructure> structures) {
    Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
    String personId = this.personService.getLinkedPersonId(authentication);
    List<StructureValidationRight> structureValidationRights = this.personService.getStructureValidationRights(personId);
    for (StructureValidationRight structureValidationRight : structureValidationRights) {
      List<String> subtypesCurrentUserCanValidate = structureValidationRight.getPublicationSubtypes().stream().map(PublicationSubtype::getResId)
              .collect(Collectors.toList());
      if (subtypesCurrentUserCanValidate.contains(publication.getSubtype().getResId())) {
        Structure structure = this.structureService.findById(structureValidationRight.getResId()).orElse(null);
        if (structure != null) {
          PublicationStructure pStructure = new PublicationStructure();
          pStructure.setStructure(structure);
          pStructure.setPublication(publication);
          pStructure.setLinkType(PublicationStructure.LinkType.VALIDATION);
          pStructure.getCreation().setWhen(OffsetDateTime.now());
          structures.add(pStructure);
          log.info("validation structure '{}' automatically added on publication '{}'", structure.getName(), publication.getResId());
          return true;
        }
      }
    }

    log.warn("Unable to automatically add a validation structure on publication '{}'", publication.getResId());
    return false;
  }

  /**
   * Manage links between a publication and their research groups according to the given list of groups: add research groups that exist in the
   * groups list but not in the publication's research groups list. Remove research groups from the publication's research groups that are not
   * in the groups list.
   *
   * @param publication
   * @param groups
   */
  private void syncResearchGroupsList(Publication publication, List<ResearchGroup> groups) {
    Set<ResearchGroup> researchGroups = publication.getResearchGroups();

    List<String> existingResIds = researchGroups.stream().map(ResourceNormalized::getResId).collect(Collectors.toList());
    List<String> resIdsToPersist = groups.stream().map(ResourceNormalized::getResId).collect(Collectors.toList());
    List<String> resIdsToAdd = resIdsToPersist.stream().filter(element -> !existingResIds.contains(element)).collect(Collectors.toList());
    List<String> resIdsToRemove = existingResIds.stream().filter(element -> !resIdsToPersist.contains(element)).collect(Collectors.toList());
    List<ResearchGroup> researchGroupsToRemove = researchGroups.stream().filter(group -> resIdsToRemove.contains(group.getResId()))
            .collect(Collectors.toList());

    // remove research groups that are not present anymore
    researchGroups.removeAll(researchGroupsToRemove);

    // add research group not in list yet
    for (String resId : resIdsToAdd) {
      // Add new research group
      ResearchGroup researchGroup = this.researchGroupService.findById(resId).orElse(null);
      if (researchGroup != null) {
        researchGroups.add(researchGroup);
      } else {
        throw new SolidifyResourceNotFoundException("No research group with resId '" + resId + "' found");
      }
    }
  }

  private void syncContributorsList(Publication publication, List<AbstractContributor> abstractContributors) {
    Set<Contributor> currentContributors = publication.getContributors();

    // remove all contributors from list
    currentContributors.clear();

    for (AbstractContributor abstractContributor : abstractContributors) {
      if (abstractContributor instanceof ContributorDTO) {
        ContributorDTO contributorDTO = (ContributorDTO) abstractContributor;

        // find or create contributor posted in metadata
        Contributor contributor = this.contributorService.findOrCreate(contributorDTO);
        currentContributors.add(contributor);
      }
    }
  }

  protected String updateDateFromFrontendToMetadataFormat(String dateStr) {
    return this.updateDateFormat(dateStr, this.frontendEnterDateFormat, this.metadataOutputDateFormat);
  }

  protected String updateDateFromMetadataToFrontendFormat(String dateStr) {
    return this.updateDateFormat(dateStr, this.metadataOutputDateFormat, this.frontendOutputDateFormat);
  }

  /**
   * Update a date from a format to another. The first source format is going to be checked with the first output format,
   * the second source Format with the second output format and so on, to avoid to check format that does not contain
   * the same info (i.e MM-yyyy with yyyy-MM-dd). They are both defined as the same order for an specific type of format.
   *
   * @param dateStr
   * @param sourceFormat
   * @param outputFormat
   * @return
   */
  private String updateDateFormat(String dateStr, String[] sourceFormat, String[] outputFormat) {
    int counter = 0;
    for (String source : sourceFormat) {
      try {
        DateTimeFormatter formatterFront = new DateTimeFormatterBuilder()
                .appendPattern(source)
                .parseDefaulting(ChronoField.DAY_OF_MONTH, 1)
                .parseDefaulting(ChronoField.MONTH_OF_YEAR, 1)
                .toFormatter();
        LocalDate localDate = LocalDate.parse(dateStr, formatterFront);
        DateTimeFormatter formatterMetadata = DateTimeFormatter.ofPattern(outputFormat[counter]);
        return localDate.format(formatterMetadata);
      } catch (DateTimeParseException e) {
        counter++;
      }
    }
    // return an incorrect date if it cannot be transformed, since returning the date as it is could have problems later on when validating
    // since it could be in a correct format for the backend but not for the frontend. (i.e date 2021-05 not possible to transform here but it's
    // valid date for metadata
    return MetadataExtractor.INCORRECT_DATE;
  }

}

