package ch.unige.aou.service.metadata.imports;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Map;
import java.util.Optional;
import java.util.stream.IntStream;

import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.DocumentFileService;
import ch.unige.aou.business.DocumentFileTypeService;
import ch.unige.aou.business.LicenseService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.publication.DocumentFile;
import ch.unige.aou.model.publication.DocumentFileType;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.settings.License;
import ch.unige.aou.model.xml.deposit.v2_4.DepositDoc;
import ch.unige.aou.service.DownloadService;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.metadata.MetadataExtractor;
import ch.unige.aou.service.rest.JournalTitleRemoteService;

@Service
public class UnpaywallImportService extends MetadataAndFileImportService {
  private static final Logger log = LoggerFactory.getLogger(UnpaywallImportService.class);

  private String unpaywallApiUrl;
  private String unpaywallEmail;

  private static final String UNPAYWALL_FILE_NAME = "-from-unpaywall.pdf";
  private static final String LICENSE_OTHER_AO = "OTHER-OA";
  private static final String LICENSE = "license";

  protected DownloadService downloadService;

  public UnpaywallImportService(AouProperties aouProperties, MessageService messageService, MetadataService metadataService,
          ContributorService contributorService, DocumentFileTypeService documentFileTypeService,
          DocumentFileService documentFileService, DownloadService downloadService, LicenseService licenseService,
          JournalTitleRemoteService journalTitleRemoteService) {
    super(aouProperties, messageService, metadataService, contributorService, documentFileService, documentFileTypeService,
            journalTitleRemoteService, licenseService);
    this.downloadService = downloadService;

    this.unpaywallApiUrl = aouProperties.getMetadata().getImports().getUnpaywall().getUrl();
    this.unpaywallEmail = aouProperties.getMetadata().getImports().getUnpaywall().getEmail();
  }

  public DepositDoc createDocumentFileFromUnpaywall(Publication publication) {
    final MetadataExtractor metadataExtractor = this.metadataService.getMetadataExtractor(publication.getMetadataVersion());
    DepositDoc depositDoc = (DepositDoc) metadataExtractor.createDepositDocObjectFromXml(publication.getMetadata());
    String doi = depositDoc.getIdentifiers().getDoi();
    if (!StringTool.isNullOrEmpty(doi)) {
      String unpaywallJson = this.getJsonFromUnpaywall(doi);
      JSONObject jsonObject = new JSONObject(unpaywallJson);

      try {
        // first check if is_oa is true to know that there is an open access copy for the resource
        if (jsonObject.get("is_oa").equals(Boolean.TRUE)) {
          String status = jsonObject.get("oa_status").toString();
          if (status.equals("green") || status.equals("gold") || status.equals("hybrid") || this.isBronzeStatus(status)) {

            String urlToDownload;
            String version;
            DocumentFile.AccessLevel accessLevel;
            License license = null;

            /**
             * Parse Unpaywall Json
             */
            if (this.isBronzeStatus(status) && jsonObject.has("has_repository_copy") && jsonObject.getBoolean("has_repository_copy")) {
              /**
               * Case of Bronze Open Access
               */
              Optional<JSONObject> firstLocationWithLicense = IntStream.range(0, jsonObject.getJSONArray("oa_locations").length())
                      .mapToObj(i -> jsonObject.getJSONArray("oa_locations").getJSONObject(i))
                      .filter(oa -> oa.has(LICENSE) && !StringTool.isNullOrEmpty(this.getStringOrNull(oa, LICENSE))).findFirst();

              JSONObject readLocation;
              if (firstLocationWithLicense.isPresent()) {
                // A location with license found
                readLocation = firstLocationWithLicense.get();
                accessLevel = DocumentFile.AccessLevel.PUBLIC;
                // in this case we use the OtherOA license by default
                license = this.getLicense(LICENSE_OTHER_AO);
              } else {
                // No location found without a license
                // We don't set any license on the file
                readLocation = jsonObject.getJSONArray("oa_locations").getJSONObject(0);
                accessLevel = DocumentFile.AccessLevel.RESTRICTED;
              }
              urlToDownload = this.getStringOrNull(readLocation, "url_for_pdf");
              version = this.getStringOrNull(readLocation, "version");

            } else if (jsonObject.has("best_oa_location")) {
              /**
               * Case of real Open Access
               */
              JSONObject bestOALocation = jsonObject.getJSONObject("best_oa_location");

              String licenseName = this.getStringOrNull(bestOALocation, LICENSE);
              license = this.getLicense(licenseName);
              if (license != null) {
                // if the license name returned by Unpaywall matches one of your license, we use it and set the
                // access to public by default
                accessLevel = DocumentFile.AccessLevel.PUBLIC;
              } else {
                // if the license doesn't match one of the OaU license, we use the OtherOA license and set the
                // access to restricted by default
                license = this.getLicense(LICENSE_OTHER_AO);
                accessLevel = DocumentFile.AccessLevel.RESTRICTED;
              }
              urlToDownload = this.getStringOrNull(bestOALocation, "url_for_pdf");
              version = this.getStringOrNull(bestOALocation, "version");
            } else {
              throw new SolidifyRuntimeException("DOI is not bronze and hasn't any best location");
            }
            if (urlToDownload == null) {
              throw new SolidifyRuntimeException("There is no url to download the file in UnpaywallService");
            }
            /**
             * Create DocumentFile
             */
            try {
              DocumentFile documentFile = new DocumentFile();
              documentFile.setPublication(publication);
              documentFile.setSourceData(new URI(urlToDownload));
              documentFile.setLicense(license);
              documentFile.setAccessLevel(accessLevel);
              documentFile.setDocumentFileType(this.getDocumentFileTypeFromUnpaywallVersion(version));
              documentFile.setMimetype("application/pdf");
              documentFile.setFileName(publication.getSubtype().getName().replace(" ", "-") + UNPAYWALL_FILE_NAME);
              documentFile.setIsImported(true);
              this.saveDocumentFileIfSourceDataValid(documentFile);
            } catch (URISyntaxException e) {
              throw new SolidifyRuntimeException("Unable to get file URL from unpaywall", e);
            }
          } else {
            log.info("The license of the file is closed");
          }
        } else {
          log.info("The doi found through Unpaywall is not in open access");
        }
      } catch (JSONException | SolidifyRuntimeException e) {
        // catch JSONEXception and SolidifyRuntimeException to not block deposit process if Unpaywall import
        // fails
        log.warn("Unable to parse Json returned by Unpaywall", e);
      }
    } else {
      log.info("There is no doi to check through Unpaywall");
    }
    return depositDoc;
  }

  private License getLicense(String licenseName) {
    if (!StringTool.isNullOrEmpty(licenseName)) {
      return this.licenseService.findById(licenseName.toUpperCase()).orElse(null);
    }
    return null;
  }

  private DocumentFileType getDocumentFileTypeFromUnpaywallVersion(String version) {
    if (!StringTool.isNullOrEmpty(version)) {
      switch (version) {
        case "publishedVersion":
          return this.documentFileTypeService.findByValue("Article (Published version)");
        case "acceptedVersion":
          return this.documentFileTypeService.findByValue("Article (Accepted version)");
        case "submittedVersion":
          return this.documentFileTypeService.findByValue("Article (Submitted version)");
        default:
          return null;
      }
    }
    return null;
  }

  private String getJsonFromUnpaywall(String doi) {
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    RestTemplate client = restTemplateBuilder.build();
    return client.getForObject(this.unpaywallApiUrl, String.class, Map.of("doi", doi, "email", this.unpaywallEmail));
  }

  private boolean isBronzeStatus(String status) {
    return status.equals("bronze");
  }
}
