package ch.unige.aou.service.rest;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.context.annotation.Lazy;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.PublicationService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.ExistingPublicationInfo;
import ch.unige.aou.model.publication.Publication;

@Service
public class DuplicateService {

  private static final String DOI = "doi";
  private static final String PUBMED = "pubmed";
  private static final String ARXIV = "arxiv";
  private static final String ISBN = "isbn";

  private static final String FOUND_MESSAGE = "this document already exists";
  private static final String NOT_FOUND_MESSAGE = "this document does not exist";

  private String identifiersCheckUrl;
  private String identifiersCheckUrlWithSubType;

  private PublicationService publicationService;

  public DuplicateService(@Lazy PublicationService publicationService, AouProperties aouProperties) {
    this.publicationService = publicationService;
    this.identifiersCheckUrl = aouProperties.getMetadata().getSearch().getDuplicates().getIdentifiersCheckUrl();
    this.identifiersCheckUrlWithSubType = aouProperties.getMetadata().getSearch().getDuplicates().getIdentifiersCheckUrlWithSubType();
  }

  /******/

  /**
   * Return true if the DOI exists in a publication that is already published or is used by another publication in database.
   *
   * @param doi
   * @param publicationId
   * @return
   */
  public boolean doiAlreadyUsedByAnotherPublication(String doi, String publicationId) {
    ExistingPublicationInfo existingPublicationInfo = this.getExistingPublicationInfoByDoi(doi);
    if (existingPublicationInfo == null) {
      return false;
    } else if (!StringTool.isNullOrEmpty(existingPublicationInfo.getResId())) {
      // check if publication is different
      return !publicationId.equals(existingPublicationInfo.getResId());
    } else {
      // publication is already online
      return true;
    }
  }

  public ExistingPublicationInfo getExistingPublicationInfoByDoi(String doi) {
    Optional<Publication> publicationOptional = this.publicationService.findByDoi(doi);
    if (publicationOptional.isPresent()) {
      return new ExistingPublicationInfo(publicationOptional.get().getResId());
    } else {
      String url = this.getUrlOfAlreadyPublishedIdentifier(DOI, doi);
      if (!StringTool.isNullOrEmpty(url)) {
        return new ExistingPublicationInfo(null, url);
      }
    }
    return null;
  }

  /******/

  /**
   * Return true if the arxivId exists in a publication that is already published.
   *
   * @param arxivId
   * @return
   */
  public boolean arxivIdAlreadyExists(String arxivId) {
    return this.getExistingPublicationInfoByArxivId(arxivId) != null;
  }

  public ExistingPublicationInfo getExistingPublicationInfoByArxivId(String arxivId) {
    String url = this.getUrlOfAlreadyPublishedIdentifier(ARXIV, arxivId);
    if (!StringTool.isNullOrEmpty(url)) {
      return new ExistingPublicationInfo(null, url);
    }
    return null;
  }

  /******/

  /**
   * Return true if the PMID exists in a publication that is already published or is used by another publication in database.
   *
   * @param pmid
   * @param publicationId
   * @return
   */
  public boolean pmidAlreadyUsedByAnotherPublication(String pmid, String publicationId) {
    ExistingPublicationInfo existingPublicationInfo = this.getExistingPublicationInfoByPmid(pmid);
    if (existingPublicationInfo == null) {
      return false;
    } else if (!StringTool.isNullOrEmpty(existingPublicationInfo.getResId())) {
      // check if publication is different
      return !publicationId.equals(existingPublicationInfo.getResId());
    } else {
      // publication is already online
      return true;
    }
  }

  public ExistingPublicationInfo getExistingPublicationInfoByPmid(String pmid) {
    Optional<Publication> publicationOptional = this.publicationService.findByPmid(pmid);
    if (publicationOptional.isPresent()) {
      return new ExistingPublicationInfo(publicationOptional.get().getResId());
    } else {
      String url = this.getUrlOfAlreadyPublishedIdentifier(PUBMED, pmid);
      if (!StringTool.isNullOrEmpty(url)) {
        return new ExistingPublicationInfo(null, url);
      }
    }
    return null;
  }

  /******/

  /**
   * Return true if the ISBN exists in a publication that is already published.
   *
   * @param isbn
   * @return
   */
  public boolean bookWithIsbnAlreadyExists(String isbn) {
    return this.getExistingPublicationInfoByIsbn(isbn) != null;
  }

  public ExistingPublicationInfo getExistingPublicationInfoByIsbn(String isbn) {
    String url = this.getUrlOfAlreadyPublishedIdentifier(ISBN, isbn);
    if (!StringTool.isNullOrEmpty(url)) {
      return new ExistingPublicationInfo(null, url);
    }
    return null;
  }

  /*************************************************/

  private String getUrlOfAlreadyPublishedIdentifier(String identifierType, String value) {
    if (!StringTool.isNullOrEmpty(this.identifiersCheckUrl) && !(StringTool.isNullOrEmpty(this.identifiersCheckUrlWithSubType))) {
      String urlToQuery = identifierType.equals(ISBN) ? this.identifiersCheckUrlWithSubType : this.identifiersCheckUrl;
      try {
        RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
        RestTemplate client = restTemplateBuilder.rootUri(urlToQuery).build();

        Map<String, String> uriVariables = new HashMap<>();
        uriVariables.put("identifierType", identifierType);
        uriVariables.put("value", value);

        if (identifierType.equals(ISBN)) {
          uriVariables.put("subtypeValue", AouConstants.ISBN_SUBTYPE_TO_CHECK_DUPLICATE);
        }

        final String jsonResponse = client.getForObject(urlToQuery, String.class, uriVariables);

        // HTTP 200 --> identifier already exists
        return this.getUrlFromFoundMessage(jsonResponse);

      } catch (final HttpClientErrorException e) {
        if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
          // HTTP 404 --> id not deposited yet
          this.checkNotFoundMessage(e.getResponseBodyAsString());
          return null;
        } else {
          throw new SolidifyRuntimeException(e.getMessage());
        }
      }
    } else {
      throw new SolidifyRuntimeException("incomplete configuration: identifiersCheckUrl is not set");
    }
  }

  private String getUrlFromFoundMessage(String jsonResponse) {
    try {
      ObjectMapper mapper = new ObjectMapper();
      JsonNode responseNode = mapper.readValue(jsonResponse, JsonNode.class);
      if (responseNode.has("success")) {
        JsonNode successNode = responseNode.get("success");
        String successMessage = successNode.get(0).asText();
        if (FOUND_MESSAGE.equals(successMessage) && responseNode.has("url")) {
          return responseNode.get("url").textValue();
        }
      }
      throw new SolidifyRuntimeException("incoherent service response");
    } catch (IOException e) {
      throw new SolidifyRuntimeException(e.getMessage());
    }
  }

  private boolean checkNotFoundMessage(String jsonResponse) {
    try {
      ObjectMapper mapper = new ObjectMapper();
      JsonNode responseNode = mapper.readValue(jsonResponse, JsonNode.class);
      if (responseNode.has("errors")) {
        JsonNode errorsNode = responseNode.get("errors");
        String errorMessage = errorsNode.get(0).asText();
        if (NOT_FOUND_MESSAGE.equals(errorMessage)) {
          return true;
        }
      }
      throw new SolidifyRuntimeException("incoherent service response");
    } catch (IOException ex) {
      throw new SolidifyRuntimeException(ex.getMessage());
    }
  }
}
